#!/usr/bin/env python
# www.byvac.com
# Contains tools for bvuP


import sys, struct, os, os.path, copy,  wx
if sys.version_info[0] == 3:
    from urllib.request import urlopen
else:
    # Not Python 3 - today, it is most likely to be Python 2
    # But note that this might need an update when Python 4
    # might be around one day
    from urllib import urlopen

from time import time, sleep
from stat import *
#if os.name == 'nt':
#    from pyperclip import winGetClipboard, winSetClipboard
#else:
#    import clipboard

# mictopython
MRAW = b'\x01'
MRAWEXIT = b'\x02'
MBREAK = b'\x03'
MRESET = b'\x04'
MRAWSEND = b'\x04'
MPASTE = b'\x05'
MPASTEXIT = b'\04'

app = wx.App(0)

# ==============================================================================
# U T I L I T I E S
# ==============================================================================
# ******************************************************************************
# upens a url or file
# ******************************************************************************
def furl_open(fname):
    f = None
    if 'http:' in fname:
        try:
            f = urlopen(fname)
        except:
            print("\r\nCant open url:")
            print( fname)
    else:
        try:
            nfname = fname.replace('\\','/')
            f = open(nfname, "rb")
        except:
            print("\r\nCant open file:",nfname)
    return f

# ******************************************************************************
# Opens a local file or a url and returns with a list of lines
# ******************************************************************************
def get_lines(fname):
    rv = []
    f = furl_open(fname)
    if f:
        line = f.readline()
        while len(line) > 0:
            ln = line.decode('utf-8') # 3 now brings back bytes
            rv.append(ln.rstrip())
            line = f.readline()
    rv.append('') # needs an extra CR at end
    return rv


# ==============================================================================
# M I C R O P Y T H O N
# ==============================================================================
# ******************************************************************************
# go into raw mode
# ******************************************************************************
def rawMode(sp):
    sp.flushInput()
    sp.write(MRAW) # mode
    waitFor(sp,'exit',500) # raw mode sign on

# ******************************************************************************
# go into raw mode
# ******************************************************************************
def rawExit(sp):
    sp.flushInput()
    sp.write(MRAWEXIT) # mode
    waitFor(sp,'information.',500) # raw mode sign off

# ******************************************************************************
# in raw mode ends block
# ******************************************************************************
def blockEnd(sp):
    sp.write(MRAWSEND) # ends block
    valid,rv=waitFor(sp,'OK',500)
    #print('at block end valid={},rv={}'.format(valid,rv))

# ******************************************************************************
# waits for string with timeout, returns true if found and all input upto
# and including
# Timeout is always im mS
# ******************************************************************************
def waitFor(sp, text, timeout):
    endat = time() + (timeout / 1000) # mS
    rv = ''
    while time() < endat:
        raw = sp.read(1)
        if len(raw) > 0:
            rv = rv + chr(raw[0])
            endat = time() + (timeout / 1000) # reset
        if text in rv:
            return True, rv  
    if time() > endat:
        return False, rv

# ******************************************************************************
# os utilities
# When ftype is 'dir' this expects to be in that dir, this is the action from
# a GUI tree view
# ******************************************************************************
def delete_file(sp,file,ftype):
    sp.write(str.encode('import os\r\n'))
    if ftype == 'file':
        sp.write(str.encode('os.remove("'+file+'")\r\n'))
    if ftype == 'dir':
        sp.write(str.encode('os.chdir("..")\r\n')) # if dir is selected then we will be in that dir
        sp.write(str.encode('os.rmdir("'+file+'")\r\n'))

# ******************************************************************************
# create new folder
# ******************************************************************************
def new_folder(sp, folder_name):
    sp.write(str.encode('import os\r\n'))
    sp.write(str.encode('os.mkdir("'+folder_name+'")\r\n'))

# ******************************************************************************
# used for writing in raw mode
# ******************************************************************************
def rawWrite(sp,text):
    sp.flushOutput()
    sp.write(str.encode(text+'\r\n'))
    #print 'writing:',text+'\r\n'
    sp.write(MRAWSEND)
    valid, rv = waitFor(sp,'OK',500)
    #print 'rv=',rv
    if valid == False:
        print("Error")
        print(rv)
    
# ******************************************************************************
# Sends a string, which is split into lines
# ******************************************************************************
def paste_send(sp,command):
#    oldto = sp.timeout
#    sp.timeout = 0.01

    # activate paste mode ===
    sp.flushInput()
    sp.write(MPASTE)
    valid, rv = waitFor(sp,'===',500)

    # send lines
    if valid:
        sp.flushInput()
        lines = command.splitlines()
        for l in lines:
            sp.write(str.encode(l+'\r\n'))
            waitFor(sp,'===',100)

    # check send ok
    sp.write(MPASTEXIT) # ends block in past mode ===
    valid, rv = waitFor(sp,'>>>',2000)
#    for l in rv.splitlines():
#        print('l=',l)
    
    if valid:
        if 'Traceback' in rv: valid = False
    return valid, rv

# ******************************************************************************
# sends simple commands or strings as raw input, don't use if indents are 
# involved
# ******************************************************************************
def replSend(sp,cmd):
    sp.flushInput()
    sp.write(str.encode(cmd))
    return waitFor(sp,'>>>',500)
            
# ******************************************************************************
# paste mode
# ******************************************************************************
def pasteMode(sp):
    sp.write(MPASTE) # paste mode
    waitFor(sp,'=== \r\n===',1000) # paset mode sign on
#    sleep(5) # past === and ===

# ==============================================================================
# D E V I C E
# ==============================================================================
# ******************************************************************************
# files needed for 
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
def check_device(sp):
    sp.flushInput()
    sp.write(b'\r\n')
    if waitFor(sp,'>>>',1000):
        return True
    else:
        print("No device")
        return False

# ******************************************************************************
# Sends contents of clipboard as if typed
# ******************************************************************************
def clip_transfer(sp):
    rv = None
    if wx.TheClipboard.Open():
        data = wx.TextDataObject()
        wx.TheClipboard.GetData(data)
        clipText = data.GetText()
        wx.TheClipboard.Close()
        #print('cliptext',clipText)
    else:
        rv = 'Unable to get contents of clipboard'
        return False, rv

    if len(clipText) == 0:
        rv = "Cipboard is empty"
        return False, rv

    return paste_send(sp,clipText) # returns valid,rv
    #print('(1)valid={}, rv={}'.format(valid,rv) )
    
# ******************************************************************************
# used in paste mode, line is echoed and followed by CR ===
# ******************************************************************************
def pasteWrite(sp,text):
    sp.flushInput()
    sp.write(str.encode(text+'\r\n'))
#    print text+'\r\n'
    valid, rv = waitFor(sp,'===',500)
    sp.flushOutput()

# ******************************************************************************
# Gets dictionary with all files by directory ;0=file ;1=dir
# {'/;1': ['boot.py;0', 'system_calls.py;0', 'test.py;0', 'bv4242.py;0',
# 'dir1;1', 'i2cn20.py;0', 'i2cn18.py;0'], 'dir1;1': ['bv112.py;0', 'boot_old.py;0',
# 'dir2;1', 'Notes.txt;0'], 'dir2;1': ['i2cn18.py;0', 'iic.py;0']}
# ******************************************************************************
def listDir(sp):
    import ast
    prog = '''
import os
def x__get_files():
	rv = []
	for n in os.ilistdir():
		if n[1] & 0x4000 !=0:
			rv.append(n[0]+';1')
		else:
			rv.append(n[0]+';0')
	return rv

dirname = '/;1'
all = {dirname:[]}
stack = []
def x__build(stack,dirname):
	os.chdir(dirname.split(';')[0])
	stack.append(x__get_files())
	#print('lenstack={},stack={}'.format(len(stack),stack))
	if len(stack) > 0:
		list = stack.pop(len(stack)-1)
		while len(list) > 0:
			#print('list=',list)
			name = list.pop(0)
			#print('name=',name)
			type = name.split(';')
			all[dirname].append(name)
			if type[1] != '0': # dir
				all[name]=[]
				x__build(stack,name)
        if dirname != '/;1':
           os.chdir('..')
x__build(stack,dirname)
os.chdir('/')
print(all)			
'''
    rv = ''
    if check_device(sp):
        valid, rv = paste_send(sp,prog)
        if valid:
            rv1 = rv.replace('>>>','') # remove prompt from end
            try:
                rv = ast.literal_eval(rv1)
            except:
                rv = ''         
    return rv

# ******************************************************************************
# change directory
# ******************************************************************************
def cd(sp,dir):
    prog = '''
import os
os.chdir(\'{}\')
'''.format(dir)    
    return paste_send(sp,prog)
    

# ******************************************************************************
# return contents of file
# ******************************************************************************
def get_file_contents(sp,fname):
    prog = """
for line in open(\'{0}\'):
   print(line,end='')    
""".format(fname)
    print(prog)
    if check_device(sp):
        valid, rv = paste_send(sp, prog)
        rv = rv[:-3] # remove >>>
    else:
        return False, ''
    return valid, rv

# ******************************************************************************
# uploads a file, no checking is done to see if it exists
# has to use binascii because of possible ' or " convert form here and back
# again
# ******************************************************************************
def saveText(sp,lines,upfile,win):
    #count =0
    #print('sending ',upfile)
    if check_device(sp):
        #print('device ok, lines=',lines)
        import binascii        
        rawMode(sp)
        sp.write(str.encode('import ubinascii\r\n'))
        #print('sending ',str.encode('import ubinascii\r\n'))
        blockEnd(sp)
        sp.write(str.encode("f = open('{0}', 'wb')\r\n".format(upfile)))
        #print('sending ',str.encode("f = open('{0}', 'wb')\r\n".format(upfile)))
        blockEnd(sp)
        for l in lines:
            l+='\r'
            #print('{} {}'.format(len(l),l))
            #count +=1
            hex = binascii.hexlify(l.encode('utf-8')) # encode to hex, requires bytes
            sp.write(b'f.write(ubinascii.unhexlify("'+hex+b'")\r\n)')
            blockEnd(sp)
            win.txtTerm.AppendText('.')
            #sleep(0.1) # required
        sp.write(str.encode("f.close()"))
        blockEnd(sp)
        rawExit(sp)
#    sp.timeout = oldto
    win.txtTerm.AppendText('\r\n')

def sendfile(sp,infile,upfile,win):
    lines = get_lines(infile)
    saveText(sp,lines,upfile,win)

# ******************************************************************************
# uploads a file, no checking is done to see if it exists
# has to use binascii because of possible ' or " convert form here and back
# again
# ******************************************************************************
def saveAny(sp,infile,upfile,win):
    #count =0
    print('sending ',upfile)
    fi = open(infile, 'rb')
    if fi is None:
        return
    if check_device(sp):
        print('device ok, infilr=',infile)
        import binascii        
        rawMode(sp)
        sp.write(str.encode('import ubinascii\r\n'))
        print('sending ',str.encode('import ubinascii\r\n'))
        blockEnd(sp)
        sp.write(str.encode("f = open('{0}', 'wb')\r\n".format(upfile)))
        print('sending ',str.encode("f = open('{0}', 'wb')\r\n".format(upfile)))
        blockEnd(sp)
        
        while True:
            chunk = fi.read(64)
            if chunk == b'':
                break
            hex = binascii.hexlify(chunk) # encode to hex
#            print('hex=',hex)
            sp.write(b'f.write(ubinascii.unhexlify("'+hex+b'")\r\n)')
            win.txtTerm.AppendText('.')
            blockEnd(sp)
        sp.write(str.encode("f.close()"))
        blockEnd(sp)
        rawExit(sp)
#    sp.timeout = oldto
    win.txtTerm.AppendText('\r\n')
# ------------ old -----------------




# ******************************************************************************
# removes a device file
# <filename>
# <*.*> all, use dos type, only tested on * 
# ******************************************************************************
def remove(sp,fn):
    if check_device(sp):
        import ast, re
        replSend(sp,'import os\r\n')
        valid, rv = replSend(sp,'os.listdir()\r\n')
        # rv now has something like
        # os.listdir()
        # ['boot.py', '__upload.py', '__download.py']
        files = ast.literal_eval(rv.splitlines()[1]) # convert list string to list
        fn = fn.replace('*','(.*)') # for regular expression
        for n in files:
            try:
                mo = re.search(fn, n, re.M|re.I)            
                if mo:
                    print('removing', mo.group())
                    sp.write(str.encode('os.remove("'+mo.group()+'")\r\n'))
            except:
                print("\r\nError in expression")
                break    

# ******************************************************************************
# used for transferring files between device and PC
# can only habdle text for now
# ******************************************************************************
def getfile(sp,infile,outfile):
    if check_device(sp):
        # get contents of infile
        rawMode(sp)
        sp.flushInput()
        sp.write(str.encode("with open('"+infile+"', 'r') as file:\r\n"))
        sp.write(str.encode("  print(file.read())\r\n"))
        sp.write(MRAWSEND) # do it
        valid, xx = waitFor(sp,'OK',500) # okay from
        valid, rv = waitFor(sp,'\04',500)
        # contents now in rv
        try:
            f = open(outfile, 'w')
            for n in rv[:-1].splitlines():
                f.write(n+'\r\n')
            f.close()
        except:
            print("problem opening ",outfile)
            
        rawExit(sp)

# ******************************************************************************
# getlocal - upload any file
# getlocal <fname> uses same name
# getlocal - wild cards
# ******************************************************************************
def getlocal(sp,imp,cf):
    if check_device(sp):
        if len(imp.split()) == 1: # just .gl
            if len(cf.filename) > 0:
                print('Getting last file', cf.filename)
                imp += ' '+cf.filename
            else:
                print('Filename required\r\n')
                return

        import re
        localfiles = os.listdir(os.getcwd())
        
        if len(imp.split()) > 2: # must be in, out file
            infile = imp.split()[1]
            outfile = imp.split()[2]
            if infile in localfiles:
                #print 'infile=',infile,'outfile=',outfile
                sendfile(sp,infile,outfile)
            else:    
                print('Filename '+infile+'\r\n')
                print('Does not exist in '+os.getcwd()+'\r\n')
                
        else:
            # may be single file name or *.py -- wile.wild
            match = imp.split()[1].replace('*','(.*)') # for regular expression
            for n in localfiles:
                try:
                    mo = re.search(match, n, re.M|re.I)            
                    if mo:
                        print('saving file', mo.group())
                        sendfile(sp,mo.group(),mo.group())
                except:
                    print("\r\nError in expression")
                    break   

# ******************************************************************************
# putlocal - upload any file
# putlocal <fname> uses same name
# putlocal - wild cards
# ******************************************************************************
def putlocal(sp,imp,cf):
    if check_device(sp):
        if len(imp.split()) == 1: # just .gl
            if len(cf.filename) > 0:
                print('Getting last file', cf.filename)
                imp += ' '+cf.filename
            else:
                print('Filename required\r\n')
                return


        import ast, re
        # get device files, cnt get if it is not there
        replSend(sp,'import os\r\n')
        valid, rv = replSend(sp,'os.listdir()\r\n')
        # rv now has something like
        # os.listdir()
        # ['boot.py', '__upload.py', '__download.py']
        dfiles = ast.literal_eval(rv.splitlines()[1]) # convert list string to list
        
        if len(imp.split()) > 2: # must be in, out file
            infile = imp.split()[1]
            outfile = imp.split()[2]
            if infile in dfiles:
                print('infile=',infile,'outfile=',outfile)
                getfile(sp,infile,outfile)
            else:    
                print('Filename '+infile+'\r\n')
                cprint('Does not exist')
                
        else:
            # may be single file name or *.py -- wile.wild
            match = imp.split()[1].replace('*','(.*)') # for regular expression
            for n in dfiles:
                try:
                    mo = re.search(match, n, re.M|re.I)            
                    if mo:
                        print('saving file', mo.group())
                        getfile(sp,mo.group(),mo.group())
                except:
                    print("\r\nError in expression")
                    break    
        
# ==============================================================================
# L O C A L (H O S T)

# ******************************************************************************
# .lcwd returns current working directory
# ******************************************************************************
def getDir():
    colour.ccprint(colour.COL_ESC, os.getcwd())

# ******************************************************************************
# .lls list files in local directory
# ******************************************************************************
def localDir():
    print('Path:', os.getcwd())
    dirs = filter(os.path.isdir, os.listdir(os.getcwd()))
    for n in dirs:
        print(n+'  <dir>')
    files = filter(os.path.isfile, os.listdir(os.getcwd()))
    for n in files:
        print(n)

# ******************************************************************************
# .lcd 'full path to new dir'
# .lcd ..  up one
# ******************************************************************************
def changeDir(newdir):
    try:
        os.chdir(newdir)
        print('Path:', os.getcwd())
    except:
        print("No such directory")
    

